// Copyright 2024 The Sigstore Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package internal

import (
	"testing"
)

func TestStripUrl(t *testing.T) {
	endpoint := "https://private.rekor.com/rekor"
	host, basePath := StripURL(endpoint)
	if host != "private.rekor.com" || basePath != "/rekor" {
		t.Fatalf("Host and/or BasePath are not correct")
	}
}
